"""
Configuration file for the Sphinx documentation builder.

For the full list of built-in configuration values, see the documentation:
https://www.sphinx-doc.org/en/master/usage/configuration.html
"""

import importlib.metadata

# Project information
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Genja"
copyright = "2025, Gavin Wiggins"
author = "Gavin Wiggins"
release = importlib.metadata.version("genja")

# General configuration
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = []

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# Options for HTML output
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_static_path = ["_static"]

html_theme = "furo"

html_theme_options = {
    "source_repository": "https://github.com/wigging/genja",
    "source_branch": "main",
    "source_directory": "docs/",
}

html_title = "Genja"



"""Function to run a local server."""

import webbrowser
from livereload import Server


def run_server(config: dict[str, str]):
    """Run a local web server.

    Run a local web server and open a browser to view the website. The website
    is automatically reloaded when changes occur in the input directory.

    Parameters
    ----------
    config
        Configuration settings from the config file.
    """
    # Get the Markdown and output directories
    output_dir = config["site_output"]

    # Open web browser to view website on localhost port
    webbrowser.open("http://localhost:5500")

    # Serve the website from the output directory and automatically reload it
    # when changes occur in the input directory
    server = Server()
    server.watch("_posts", "genja build")
    server.serve(root=output_dir)



"""Module for build functions."""

import json
import markdown
import textwrap

from bs4 import BeautifulSoup
from datetime import datetime
from jinja2 import Environment, FileSystemLoader, Template
from operator import itemgetter
from pathlib import Path


def _build_posts(
    config: dict[str, str], template: Template, mdown: markdown.Markdown
) -> list[dict[str, str]]:
    """Build HTML for Markdown files that are located in the _posts directory.

    Parameters
    ----------
    config
        Configuration settings from the config file.
    template
        Jinja template.
    mdown
        Markdown content.

    Returns
    -------
    posts
        List of dictionaries that contain post metadata.
    """
    base_url = config["base_url"]
    site_output = config["site_output"]
    posts_output = config["posts_output"]

    # Store meta data dictionary for each post
    posts = []

    for path in Path("_posts").glob("**/*.md"):
        # Get category, link, url, and post path
        category = path.parts[-2]

        if category != "_posts":
            link = f"{posts_output}/{category}/{path.name}".replace("md", "html")
            url = f"{base_url}/{posts_output}/{category}/{path.name}".replace("md", "html")
            post_path = Path(f"{site_output}/{posts_output}/{category}/{path.name}")
        else:
            link = f"{posts_output}/{path.name}".replace("md", "html")
            url = f"{base_url}/{posts_output}/{path.name}".replace("md", "html")
            post_path = Path(f"{site_output}/{posts_output}/{path.name}")

        # Get HTML string for the JSON feed
        with path.open() as f:
            mdtext = f.read()

        html = mdown.convert(mdtext)
        soup = BeautifulSoup(html, "html.parser")
        html_str = json.dumps(str(soup.p) + f'<p><a href="{url}">Continue reading...</a></p>')

        # Get meta data from the Markdown file
        meta = mdown.Meta  # pyright: ignore
        title = meta["title"][0]
        long_date = meta["date"][0]
        iso_date = datetime.strptime(meta["date"][0], "%B %d, %Y").isoformat() + "Z"

        # Store the meta data dictionary for the post
        meta_data = {
            "title": title,
            "date": long_date,
            "category": category,
            "link": link,
            "url": url,
            "iso_date": iso_date,
            "html": html_str,
        }

        posts.append(meta_data)

        # Render the post template then write to HTML file
        post_html = template.render(meta=meta_data, content=html)

        post_path = post_path.with_suffix(".html")
        post_path.parent.mkdir(parents=True, exist_ok=True)

        with post_path.open("w") as f:
            f.write(post_html)

        # Reset the Markdown parser
        mdown.reset()

    return posts


def _build_pages(config: dict[str, str], template: Template, mdown: markdown.Markdown):
    """Build HTML for Markdown files that are located in the _pages directory.

    Parameters
    ----------
    config
        Configuration settings from the config file.
    template
        Jinja template.
    mdown
        Markdown content.
    """
    site_output = config["site_output"]

    for path in Path("_pages").glob("**/*.md"):
        # Read text content of the Markdown file
        with path.open() as f:
            mdtext = f.read()

        # Convert Markdown content to HTML content
        html = mdown.convert(mdtext)

        # Create new HTML path for the page
        page_path = Path(f"{site_output}/{path.name}")
        page_path = page_path.with_suffix(".html")

        # Render the page template then write to HTML file
        page_html = template.render(content=html)
        page_path.parent.mkdir(parents=True, exist_ok=True)

        with page_path.open("w") as f:
            f.write(page_html)

        # Reset the Markdown parser
        mdown.reset()


def _build_templates(
    config: dict[str, str], templates: list[Template], names: list[str], posts: list[dict[str, str]]
):
    """Build HTML for certain templates located in the _templates directory.

    Parameters
    ----------
    config
        Configuration settings from the config file.
    templates
        List of Jinja templates.
    names
        List of names for the generated HTML pages.
    posts
        List of dictionaries that contain post metadata.
    """
    site_output = config["site_output"]

    # Render the page templates and write to HTML files
    for template, name in zip(templates, names):
        page_html = template.render(posts=posts)
        page_path = Path(f"{site_output}/{name}")

        with page_path.open("w") as f:
            f.write(page_html)


def _build_feed(config: dict[str, str], posts: list[dict[str, str]]):
    """Build the JSON feed for the posts.

    Parameters
    ----------
    config
        Configuration settings from the config file.
    posts
        List of dictionaries that contain post metadata.
    """
    base_url = config["base_url"]
    site_output = config["site_output"]
    title = config["title"]

    # Sort feed dictionaries using date
    sorted_posts = sorted(posts, key=itemgetter("iso_date"), reverse=True)

    # Templates for building JSON feed
    json_start = textwrap.dedent(f"""\
    {{
        "version": "https://jsonfeed.org/version/1.1",
        "title": "{title}",
        "home_page_url": "{base_url}",
        "feed_url": "{base_url}/feed.json",
        "items": [""")

    json_end = textwrap.dedent("""
        ]
    }""")

    # Build the JSON feed and write to a JSON file
    content = ""

    for post in sorted_posts:
        content += f"""
        {{
            "id": "{post["url"]}",
            "url": "{post["url"]}",
            "title": "{post["title"]}",
            "date_published": "{post["iso_date"]}",
            "content_html": {post["html"]}
        }},"""

    json_feed = json_start + content[:-1] + json_end
    feed_path = Path(f"{site_output}/feed.json")

    with feed_path.open("w") as f:
        f.write(json_feed)


def build_website(config: dict[str, str]):
    """Build the website.

    Parameters
    ----------
    config
        Configuration settings from the config file.
    """
    # Setup the Markdown converter
    md = markdown.Markdown(extensions=["meta", "fenced_code"])

    # Setup the jinja template environment
    loader = FileSystemLoader("_templates")
    env = Environment(loader=loader, trim_blocks=True, lstrip_blocks=True)

    # Build the posts
    post_template = env.get_template("post.html")
    posts = _build_posts(config, post_template, md)

    # Build the pages if they exist
    if Path("_templates/page.html").exists() and Path("_pages").exists():
        page_template = env.get_template("page.html")
        _build_pages(config, page_template, md)

    # Build certain templates as pages too
    page_names = []
    page_templates = []

    for f in Path("_templates").glob("*.html"):
        if f.name != "post.html" and f.name != "page.html" and f.name != "base.html":
            page_template = env.get_template(f.name)
            page_templates.append(page_template)
            page_names.append(f.name)

    _build_templates(config, page_templates, page_names, posts)

    # Build the JSON feed
    _build_feed(config, posts)

    print(f"\nBuilt website in `{config['site_output']}` directory.")



"""Static site generator for GitHub Pages."""

import argparse
import tomllib
import importlib.resources
import shutil
from importlib.metadata import version
from pathlib import Path

from .builder import build_website
from .server import run_server
from . import static


def create_project():
    """Create a new example project."""
    static_dir = importlib.resources.files(static)

    # Copy config file
    shutil.copy(f"{static_dir / 'genja.toml'}", ".")

    # Copy template files
    Path("./_templates").mkdir(exist_ok=True)
    shutil.copy(f"{static_dir / 'index.html'}", "./_templates")
    shutil.copy(f"{static_dir / 'page.html'}", "./_templates")
    shutil.copy(f"{static_dir / 'post.html'}", "./_templates")

    # Copy posts files
    Path("./_posts/fruits").mkdir(parents=True, exist_ok=True)
    shutil.copy(f"{static_dir / 'apple.md'}", "./_posts/fruits")
    shutil.copy(f"{static_dir / 'orange.md'}", "./_posts/fruits")

    Path("./_posts/veggies").mkdir(parents=True, exist_ok=True)
    shutil.copy(f"{static_dir / 'broccoli.md'}", "./_posts/veggies")
    shutil.copy(f"{static_dir / 'spinach.md'}", "./_posts/veggies")

    # Copy pages files
    Path("./_pages").mkdir(exist_ok=True)
    shutil.copy(f"{static_dir / 'about.md'}", "./_pages")
    shutil.copy(f"{static_dir / 'contact.md'}", "./_pages")

    # Copy other files
    Path("./mysite/img").mkdir(parents=True, exist_ok=True)
    shutil.copy(f"{static_dir / 'apple.jpg'}", "./mysite/img")
    shutil.copy(f"{static_dir / 'styles.css'}", "./mysite")


def remove_files(config: dict[str, str]):
    """Remove the generated HTML and JSON feed files."""
    pages_path = Path("_pages")
    posts_path = Path("_posts")
    templates_path = Path("_templates")
    output_path = Path(config["site_output"])
    blog_path = output_path / Path(config["posts_output"])

    # HTML files generated from Markdown files in pages directory
    p = pages_path.glob("**/*.md")
    html_files = [x.name.replace("md", "html") for x in p if x.is_file()]

    # HTML files generated from Markdown files in posts directory
    p = posts_path.glob("**/*.md")
    html_files = html_files + [x.name.replace("md", "html") for x in p if x.is_file()]

    # HTML files generated from templates
    p = templates_path.glob("**/*.html")
    for x in p:
        if x.name != "post.html" and x.name != "page.html" and x.name != "base.html":
            html_files.append(x.name)

    # Remove the generated JSON feed file in the output directory
    json_path = Path(output_path / "feed.json")

    if json_path.exists():
        json_path.unlink()

    # Remove the generated HTML files in the output directory
    for html_path in output_path.glob("**/*.html"):
        if html_path.name in html_files and html_path.parent != templates_path:
            html_path.unlink()

    # Remove empty sub-directories
    for subdir in blog_path.glob("**/*"):
        if subdir.is_dir() and not any(subdir.iterdir()):
            subdir.rmdir()

    # Remove empty blog directory
    if blog_path.exists():
        blog_path.rmdir()

    print(f"\nRemoved generated HTML files and JSON feed file in `{output_path}` directory.")


def main():
    """Run the genja program."""
    # Command line arguments
    parser = argparse.ArgumentParser(description="Genja static site generator for GitHub Pages")

    choices = ["build", "serve", "clean", "new"]
    parser.add_argument("command", choices=choices, help="genja commands")

    parser.add_argument("-v", "--version", action="version", version=version("genja"))
    args = parser.parse_args()

    # Create an example project
    if args.command == "new":
        create_project()

    # Get configuration from TOML file, requires Python 3.11 or higher
    with open("genja.toml", "rb") as f:
        config = tomllib.load(f)

    print(f"\n{'Genja command ':.<30} {args.command}")
    print(f"{'Base URL ':.<30} {config['base_url']}")
    print(f"{'Posts output directory ':.<30} {config['posts_output']}")
    print(f"{'Site output directory ':.<30} {config['site_output']}")
    print(f"{'Website title ':.<30} {config['title']}")

    # Build the website
    if args.command == "build":
        build_website(config)

    # Build the website then run a local server and open web browser
    if args.command == "serve":
        build_website(config)
        run_server(config)

    # Remove generated files in the output directory
    if args.command == "clean":
        remove_files(config)



"""
Tests for the command line interface (CLI).
"""

import subprocess
from importlib.metadata import version


def test_inputs():
    """
    Test for error if no inputs are given.
    """
    capture = subprocess.run("genja", shell=True, capture_output=True)
    returncode = capture.returncode
    assert returncode == 2


def test_version():
    """
    Test the version number.

    Compare the command line version number to the installed package version
    number. Both should give the same version number.
    """
    capture = subprocess.run("genja --version", shell=True, capture_output=True, text=True)
    stdout = capture.stdout
    assert stdout == version("genja") + "\n"
